@@ -20,6 +20,9 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
+#ifndef _ICECREAM_TEMPFILE_H_
+#define _ICECREAM_TEMPFILE_H_
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -32,3 +35,5 @@ extern "C" {
 #ifdef __cplusplus
 }
 #endif
+
+#endif
